from __future__ import annotations

import itertools
import re
from pathlib import Path
from typing import TYPE_CHECKING

from docutils import nodes
from packaging.version import Version
from sphinx.util.docutils import SphinxDirective

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from typing import ClassVar

    from myst_parser.mdit_to_docutils.base import DocutilsRenderer
    from sphinx.application import Sphinx


FULL_VERSION_RE = re.compile(r"^(\d+)\.(\d+)\.(\d+)(?:\.dev.*)?$")


class ReleaseNotes(SphinxDirective):
    required_arguments: ClassVar = 1

    def run(self) -> Sequence[nodes.Node]:
        dir_ = Path(self.arguments[0])
        # resolve relative dir
        if not dir_.is_absolute():
            src_file = Path(self.get_source_info()[0])
            if not src_file.is_file():
                msg = f"Cannot find relative path to: {src_file}"
                raise self.error(msg)
            dir_ = src_file.parent / self.arguments[0]
        if not dir_.is_dir():
            msg = f"Not a directory: {dir_}"
            raise self.error(msg)

        versions = sorted(
            (
                (Version(f.stem), f)
                for f in dir_.iterdir()
                if FULL_VERSION_RE.match(f.stem)
            ),
            reverse=True,  # descending
        )
        version_groups = itertools.groupby(
            versions, key=lambda vf: (vf[0].major, vf[0].minor)
        )
        for (major, minor), versions in version_groups:
            self.render_version_group(major, minor, versions)
        return []

    def render_version_group(
        self, major: int, minor: int, versions: Iterable[tuple[Version, Path]]
    ) -> None:
        target = nodes.target(
            ids=[f"v{major}-{minor}"],
            names=[f"v{major}.{minor}"],
        )
        section = nodes.section(
            "",
            nodes.title("", f"Version {major}.{minor}"),
            ids=[],
            names=[f"version {major}.{minor}"],
        )
        self.state.document.note_implicit_target(section)
        self.state.document.note_explicit_target(target)
        # append target and section to parent
        self.renderer.current_node.append(target)
        self.renderer.update_section_level_state(section, 2)
        # append children to section
        with self.renderer.current_node_context(section):
            for _, p in versions:
                self.render_include(p)

    def render_include(self, path: Path) -> None:
        # hacky solution because of https://github.com/executablebooks/MyST-Parser/issues/967
        from docutils.parsers.rst.directives.misc import Include
        from myst_parser.mocking import MockIncludeDirective

        srcfile, lineno = self.get_source_info()
        parent_dir = Path(srcfile).parent

        d = MockIncludeDirective(
            renderer=self.renderer,
            name=type(self).__name__,
            klass=Include,  # type: ignore  # wrong type hint
            arguments=[str(path.relative_to(parent_dir))],
            options={},
            body=[],
            lineno=lineno,
        )
        d.run()

        # TODO: replace the above with this once the above mentioned bug is fixed
        # from sphinx.util.parsing import nested_parse_to_nodes
        # return nested_parse_to_nodes(
        #    self.state,
        #    path.read_text(),
        #    source=str(path),
        #    offset=self.content_offset,
        # )

    @property
    def renderer(self) -> DocutilsRenderer:
        return self.state._renderer


def setup(app: Sphinx) -> None:
    app.add_directive("release-notes", ReleaseNotes)



from __future__ import annotations

from typing import TYPE_CHECKING

from sphinx.ext.autodoc import ObjectMember
from sphinx_toolbox.more_autodoc.autoprotocol import ProtocolDocumenter

if TYPE_CHECKING:
    from typing import Self

    from sphinx.application import Sphinx


def patch_sphinx_toolbox_autoprotocol():
    """Compat hack: https://github.com/sphinx-toolbox/sphinx-toolbox/issues/168"""

    class ObjectMemberCompat(ObjectMember):
        @classmethod
        def from_other(cls, other: ObjectMember) -> Self:
            return cls(
                other.__name__,
                other.object,
                docstring=other.docstring,
                class_=other.class_,
                skipped=other.skipped,
            )

        def __iter__(self):
            return iter([self.__name__, self.object])

    filter_orig = ProtocolDocumenter.filter_members

    def filter_members(
        self, members: list[ObjectMember], want_all: bool
    ) -> list[tuple[str, object, bool]]:
        member_tuples = [ObjectMemberCompat.from_other(m) for m in members]
        return filter_orig(self, member_tuples, want_all)

    ProtocolDocumenter.filter_members = filter_members


def setup(_app: Sphinx) -> None:
    patch_sphinx_toolbox_autoprotocol()



"""\
This file contains tests for deprecated functions.

This includes correct behaviour as well as throwing warnings.
"""

from __future__ import annotations

import warnings

import h5py
import numpy as np
import pytest
from scipy import sparse

import anndata as ad
import anndata.experimental
from anndata import AnnData
from anndata.tests.helpers import assert_equal


@pytest.fixture
def adata():
    adata = AnnData(
        X=sparse.csr_matrix([[0, 2, 3], [0, 5, 6]], dtype=np.float32),
        obs=dict(obs_names=["s1", "s2"], anno1=["c1", "c2"]),
        var=dict(var_names=["a", "b", "c"]),
    )
    adata.raw = adata
    adata.layers["x2"] = adata.X * 2
    adata.var["anno2"] = ["p1", "p2", "p3"]
    adata.X = adata.X / 2
    return adata


def test_get_obsvar_array_warn(adata):
    with pytest.warns(DeprecationWarning):
        adata._get_obs_array("a")
    with pytest.warns(DeprecationWarning):
        adata._get_var_array("s1")


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_get_obsvar_array(adata):
    assert np.allclose(adata._get_obs_array("a"), adata.obs_vector("a"))
    assert np.allclose(
        adata._get_obs_array("a", layer="x2"),
        adata.obs_vector("a", layer="x2"),
    )
    assert np.allclose(
        adata._get_obs_array("a", use_raw=True), adata.raw.obs_vector("a")
    )
    assert np.allclose(adata._get_var_array("s1"), adata.var_vector("s1"))
    assert np.allclose(
        adata._get_var_array("s1", layer="x2"),
        adata.var_vector("s1", layer="x2"),
    )
    assert np.allclose(
        adata._get_var_array("s1", use_raw=True), adata.raw.var_vector("s1")
    )


def test_obsvar_vector_Xlayer(adata):
    with pytest.warns(FutureWarning):
        adata.var_vector("s1", layer="X")
    with pytest.warns(FutureWarning):
        adata.obs_vector("a", layer="X")

    adata = adata.copy()
    adata.layers["X"] = adata.X * 3

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        adata.var_vector("s1", layer="X")
        adata.obs_vector("a", layer="X")


# This should break in 0.9
def test_dtype_warning():
    # Tests a warning is thrown
    with pytest.warns(FutureWarning):
        a = AnnData(np.ones((3, 3)), dtype=np.float32)
    assert a.X.dtype == np.float32

    # This shouldn't warn, shouldn't copy
    with warnings.catch_warnings(record=True) as record:
        b_X = np.ones((3, 3), dtype=np.float64)
        b = AnnData(b_X)
        assert not record
    assert b_X is b.X
    assert b.X.dtype == np.float64

    # Should warn, should copy
    with pytest.warns(FutureWarning):
        c_X = np.ones((3, 3), dtype=np.float32)
        c = AnnData(c_X, dtype=np.float64)
        assert not record
    assert c_X is not c.X
    assert c.X.dtype == np.float64


def test_deprecated_write_attribute(tmp_path):
    pth = tmp_path / "file.h5"
    A = np.random.randn(20, 10)
    from anndata import read_elem
    from anndata._io.utils import read_attribute, write_attribute

    with h5py.File(pth, "w") as f:
        with pytest.warns(DeprecationWarning, match=r"write_elem"):
            write_attribute(f, "written_attribute", A)

    with h5py.File(pth, "r") as f:
        elem_A = read_elem(f["written_attribute"])
        with pytest.warns(DeprecationWarning, match=r"read_elem"):
            attribute_A = read_attribute(f["written_attribute"])

        assert_equal(elem_A, attribute_A)
        assert_equal(A, attribute_A)


def test_deprecated_read(tmp_path):
    memory = AnnData(np.random.randn(20, 10))
    memory.write_h5ad(tmp_path / "file.h5ad")

    with pytest.warns(FutureWarning, match=r"`anndata.read` is deprecated"):
        from_disk = ad.read(tmp_path / "file.h5ad")

    assert_equal(memory, from_disk)


@pytest.mark.parametrize(
    ("old_name", "new_name"), anndata.experimental._DEPRECATED.items()
)
def test_warn_on_import_from_experimental(old_name: str, new_name: str):
    with pytest.warns(FutureWarning, match=rf"Importing {old_name}.*is deprecated"):
        getattr(anndata.experimental, old_name)



from __future__ import annotations

import warnings
from types import MappingProxyType
from typing import TYPE_CHECKING

import anndata

from .._io.specs import IOSpec, read_elem_as_dask
from .._types import Read, ReadCallback, StorageType, Write, WriteCallback
from ._dispatch_io import read_dispatched, write_dispatched
from .merge import concat_on_disk
from .multi_files import AnnCollection
from .pytorch import AnnLoader

if TYPE_CHECKING:
    from typing import Any


# Map old name in `anndata.experimental` to new name in `anndata`
_DEPRECATED = MappingProxyType(
    dict(
        (kv if isinstance(kv, tuple) else (kv, kv))
        for kv in (
            ("CSRDataset", "abc.CSRDataset"),
            ("CSCDataset", "abc.CSCDataset"),
            "sparse_dataset",
            "read_elem",
            "write_elem",
            ("RWAble", "typing.AxisStorable"),
            ("InMemoryElem", "typing.RWAble"),
        )
    )
)


def __getattr__(attr_name: str) -> Any:
    if new_path := _DEPRECATED.get(attr_name):
        msg = (
            f"Importing {attr_name} from `anndata.experimental` is deprecated. "
            f"Import anndata.{new_path} instead."
        )
        warnings.warn(msg, FutureWarning)
        # hacky import_object_by_name, but we test all these
        mod = anndata
        while "." in new_path:
            mod_name, new_path = new_path.split(".", 1)
            mod = getattr(mod, mod_name)
        return getattr(mod, new_path)
    msg = f"module {__name__!r} has no attribute {attr_name!r}"
    raise AttributeError(msg)


__all__ = [
    "AnnCollection",
    "AnnLoader",
    "read_elem_as_dask",
    "read_dispatched",
    "write_dispatched",
    "IOSpec",
    "concat_on_disk",
    "Read",
    "Write",
    "ReadCallback",
    "WriteCallback",
    "StorageType",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Union

import numpy as np
import pandas as pd
from numpy import ma
from scipy import sparse

from . import abc
from ._core.anndata import AnnData
from .compat import (
    AwkArray,
    CupyArray,
    CupySparseMatrix,
    DaskArray,
    H5Array,
    SpArray,
    ZappyArray,
    ZarrArray,
)

if TYPE_CHECKING:
    from typing import TypeAlias


__all__ = ["RWAble", "AxisStorable"]


ArrayDataStructureType: TypeAlias = Union[
    np.ndarray,
    ma.MaskedArray,
    sparse.csr_matrix,
    sparse.csc_matrix,
    SpArray,
    AwkArray,
    H5Array,
    ZarrArray,
    ZappyArray,
    abc.CSRDataset,
    abc.CSCDataset,
    DaskArray,
    CupyArray,
    CupySparseMatrix,
]


InMemoryArrayOrScalarType: TypeAlias = Union[
    pd.DataFrame, np.number, str, ArrayDataStructureType
]


AxisStorable: TypeAlias = Union[
    InMemoryArrayOrScalarType, dict[str, "AxisStorable"], list["AxisStorable"]
]
"""A serializable object, excluding :class:`anndata.AnnData` objects i.e., something that can be stored in `uns` or `obsm`."""

RWAble: TypeAlias = Union[
    AxisStorable,
    AnnData,
    pd.Categorical,
    pd.api.extensions.ExtensionArray,
]
"""A superset of :type:`anndata.typing.AxisStorable` (i.e., including :class:`anndata.AnnData`) which is everything can be read/written by :func:`anndata.read_elem` and :func:`anndata.write_elem`."""
